/**************************************
 * Direct video memory display driver *
 * Saku Airila 1996                   *
 **************************************/

#include<linuxmt/types.h>
#include<linuxmt/config.h>
#include<linuxmt/errno.h>
#include<linuxmt/fcntl.h>
#include<linuxmt/fs.h>

#ifdef CONFIG_CONSOLE_DIRECT

#define A_DEFAULT 0x07
#define A_BOLD 0x08
#define A_BLINK 0x80

static struct
{
   int w, h;
   int cx, cy;
   int CCBase;
   int InCmd;
   unsigned VideoSeg;
   unsigned CursorOfs;
   int CmdNdx;
   char CmdBuf[ 32 ];
   unsigned char Attr;
} Scr;

static void ScrollUp();
static void PositionCursor();
static void WriteChar();
static void CommandChar();
static void ClearRange();
static int IsCmdEnd();
static int GetParm();

extern int peekb();
extern int peekw();
extern void pokeb();
extern void pokew();
extern void outb();

void Console_gotoxy();  /* Does this need be exported? */

/* This will probably go away, but currently printk() uses it */
void con_charout( Ch )
char Ch;
{
   WriteChar( Ch );
/*   PositionCursor( Scr.cx, Scr.cy );*/
}

void WriteChar( Ch )
char Ch;
{
   if( Scr.InCmd )
   {
      CommandChar( Ch );
      return;
   }
   if( Ch == 27 )
   {
      Scr.InCmd = 1;
      return;
   }
   if( Ch == '\n' )
      Scr.cy++;
   else if( Ch == '\r' )
      Scr.cx = 0;
   else
   {
      Scr.CursorOfs = ( Scr.cx << 1 ) + ( ( Scr.w * Scr.cy ) << 1 );
      pokeb( Scr.VideoSeg, Scr.CursorOfs, Ch );
      pokeb( Scr.VideoSeg, Scr.CursorOfs + 1, Scr.Attr );
      Scr.cx++;
   }
   if( Scr.cx == Scr.w )
   {
      Scr.cx = 0;
      Scr.cy++;
   }
   if( Scr.cy == Scr.h )
   {
      ScrollUp( 1, Scr.h );
      Scr.cy--;
   }
}

/* Expects to be atomic, called only from WriteChar() */
void ScrollUp( st, en )
int st, en;
{
unsigned rdofs, wrofs;
int cnt;
   if( st > en || st < 1 )
      return;
   rdofs = ( Scr.w << 1 ) * st;
   wrofs = rdofs - ( Scr.w << 1 );
   cnt = ( Scr.w * ( en - st ) );
   for( ; cnt--; wrofs += 2, rdofs += 2 )
      pokew( Scr.VideoSeg, wrofs, peekw( Scr.VideoSeg, rdofs ) );
}

void PositionCursor( x, y )
int x, y;
{
int Pos;
   Pos = x + Scr.w * y;
   outb( 14, Scr.CCBase );
   outb( ( unsigned char )( ( Pos >> 8 ) & 0xFF ), Scr.CCBase + 1 );
   outb( 15, Scr.CCBase );
   outb( ( unsigned char )( Pos & 0xFF ), Scr.CCBase + 1 );
}

void ClearRange( x, y, xx, yy )
int x, y, xx, yy;
{
unsigned st, en, ClrW;
   st = ( x << 1 ) + y * ( Scr.w << 1 );
   en = ( xx << 1 ) + yy * ( Scr.w << 1 );
   ClrW = A_DEFAULT << 8;
   for( ; st < en; st += 2 )
      pokew( Scr.VideoSeg, st, ClrW );
}

void CommandChar( Ch )
char Ch;
{
char * p;
int t1, t2;
   Scr.CmdBuf[ Scr.CmdNdx++ ] = Ch;
   if( !IsCmdEnd( Ch ) )
      return;
   Scr.InCmd = 0;
   if( Scr.CmdBuf[ 0 ] != '[' )
   {
      Console_write( NULL, NULL, Scr.CmdBuf, Scr.CmdNdx - 1 );
      Scr.CmdNdx = 0;
      return;
   }
   switch( Ch )
   {
      case 'J' :
         if( Scr.CmdBuf[ 1 ] == '2' )
            ClearRange( 0, 0, Scr.w, Scr.h );
         break;
      case 'm' :
         switch( Scr.CmdBuf[ 1 ] )
         {
            case 'm' :
               Scr.Attr = A_DEFAULT;
               break;
            case '1' :
               Scr.Attr |= A_BOLD;
               break;
            case '5' :
               Scr.Attr |= A_BLINK;
               break;
            default :
               break;
         }
         break;
      case 'H' :
         if( Scr.CmdBuf[ 1 ] == 'H' )
         {
            Scr.cx = Scr.cy = 0;
            break;
         }
         t2 = GetParm( Scr.CmdBuf + 1 ) - 1;
         p = Scr.CmdBuf;
         while( *p != ';' )
            p++;
         t1 = GetParm( p + 1 ) - 1;
         Console_gotoxy( t1, t2 );
         break;
      case 'A' : /* up */
         t1 = GetParm( Scr.CmdBuf + 1 );
         Console_gotoxy( Scr.cx, Scr.cy - t1 );
         break;
      case 'B' : /* down */
         t1 = GetParm( Scr.CmdBuf + 1 );
         Console_gotoxy( Scr.cx, Scr.cy + t1 );
         break;
      case 'C' : /* right */
         t1 = GetParm( Scr.CmdBuf + 1 );
         Console_gotoxy( Scr.cx + t1, Scr.cy );
         break;
      case 'D' : /* left */
         t1 = GetParm( Scr.CmdBuf + 1 );
         Console_gotoxy( Scr.cx - t1, Scr.cy );
         break;
      default :
         break;
   }
   Scr.CmdNdx = 0;
}

int IsCmdEnd( Ch )
char Ch;
{
   switch( Ch )
   {
      case 27 :
      case 'J' :
      case 'm' :
      case 'H' :
         return 1;
      default :
         break;
   }
return 0;
}

int GetParm( Str )
char * Str;
{
int val = 0;
char * p = Str;
   if( *p == '-' || *p == '+' )
      p++;
   for( ; ( *p <= '9' && *p >= '0' ) ; p++ )
   {
      val *= 10;
      val += *p - '0';
   }
return *Str == '-' ? -val : val;
}

void Console_gotoxy( x, y )
int x, y;
{
   if( x >= Scr.w || y >= Scr.h )
      return;
   if( x < 0 )
      x = 0;
   if( y < 0 )
      y = 0;
/* Correct to set <0 to ==0 ? */
   Scr.cx = x;
   Scr.cy = y;
   PositionCursor( x, y );
}

int Console_lseek( inode, file, offset, origin )
struct inode *inode;
struct file *file;
off_t offset;
int origin;
{
return -ESPIPE;
}

int Console_write( inode, file, buf, count )
struct inode *inode;
struct file *file;
char *buf;
int count;
{
int cnt;
   for( cnt = 0; cnt < count; cnt++ )
      WriteChar( *buf++ );
   PositionCursor( Scr.cx, Scr.cy );
return cnt;
}

int Console_read( inode, file, buf, count )
struct inode *inode;
struct file *file;
char *buf;
int count;
{
   if( file->f_flags & O_NONBLOCK )
      return -EWOULDBLOCK;
return -EINVAL;
}

void Console_release( inode, file )
struct inode *inode;
struct file *file;
{
return;
}

int Console_open( inode, file )
struct inode *inode;
struct file *file;
{
int minor = MINOR(inode->i_rdev);
   if( minor )
      return -ENODEV;
return 0;
}

int Console_readdir()
{
return -ENOTDIR; /* Is this ok, or should it be -EBADF ? */
}

/*
 *	Once we have tty I/O in place these can go.
 */
 
static struct file_operations console_ops=
{
   Console_lseek,
   Console_read,
   Console_write,
   Console_readdir, /* readdir */
   NULL,            /* Write select later */
   NULL,            /* Write ioctl later */
   Console_open,    /* Open the bios console */
   Console_release, /* Close the bios console */
   NULL,            /* Fsync */
   NULL,            /* No media */
   NULL             /* No media */
};

void init_console()
{
   Scr.w = peekb( 0x40, 0x4a );
   Scr.h = peekb( 0x40, 0x84 );
   Scr.CCBase = peekw( 0x40, 0x63 );
   if( peekb( 0x40, 0x49 ) == 7 )
      Scr.VideoSeg = 0xB000;
   else
      Scr.VideoSeg = 0xB800;
   Scr.cx = Scr.cy = 0;
   Scr.CursorOfs = 0;
   Scr.InCmd = 0;
   Scr.CmdNdx = 0;
   Scr.Attr = A_DEFAULT;
   printk(
      "Console: Direct %dx%d Mem@%x CrtC@%x\n",
      Scr.w,
      Scr.h,
      Scr.VideoSeg,
      Scr.CCBase );
}


#endif
