#include <linuxmt/types.h>
#include <linuxmt/sched.h>
#include <linuxmt/kernel.h>
#include <linuxmt/major.h>
#include <linuxmt/string.h>
/*#include <linux/locks.h>*/
#include <linuxmt/errno.h>

#include <arch/system.h>
#include <arch/segment.h>
#include <arch/io.h>
#include <arch/irq.h>

/*
 *	STUBS for the buffer cache when we put it in
 */
 
void refile_buffer(){;}
void set_writetime(){;}

struct buffer_head *bh_chain=NULL;
struct buffer_head *bh_lru=NULL;
struct buffer_head *bh_llru=NULL;

struct wait_queue *bufwait;		/* Wait for a free buffer */

struct buffer_head buffers[NR_BUFFERS];
char bufmem[NR_BUFFERS][BLOCK_SIZE];

/*
 *	Wait on a buffer
 */
 
void wait_on_buffer(bh)
struct buffer_head * bh;
{
	struct wait_queue wait;
	
	wait.task = current;
	wait.next = NULL;

	bh->b_count++;
	add_wait_queue(&bh->b_wait, &wait);
repeat:
	current->state = TASK_UNINTERRUPTIBLE;
	if (buffer_locked(bh)) {
		schedule();
		goto repeat;
	}
	remove_wait_queue(&bh->b_wait, &wait);
	bh->b_count--;
	current->state = TASK_RUNNING;
}

void lock_buffer(bh)
struct buffer_head *bh;
{
	if(set_bit(BH_Lock, &bh->b_state))
		wait_on_buffer(bh);
}


void put_last_lru(bh)
struct buffer_head *bh;
{
	if(bh_llru==bh)
		return;
	/*
	 *	Unhook
	 */
	if(bh->b_prev_lru)
		bh->b_prev_lru->b_next_lru=bh->b_next_lru;
	if(bh->b_next_lru)
		bh->b_next_lru->b_prev_lru=bh->b_prev_lru;
	/*
	 *	Alter head
	 */
	if(bh==bh_lru)
		bh_lru=bh->b_next_lru;
	/*
	 *	Put on lru end
	 */
	bh->b_next_lru=NULL;
	bh->b_prev_lru=bh_llru;
	bh_llru->b_next_lru=bh;
	bh_llru=bh;
}

static void sync_buffers(dev,wait)
kdev_t dev;
int wait;
{
	struct buffer_head * bh;

	for(bh=bh_chain;bh!=NULL;bh=bh->b_next)
	{
		if (dev && bh->b_dev != dev)
			continue;
		/*
		 *	Skip clean buffers.
		 */
		if(!buffer_dirty(bh))
			continue;
		/*
		 *	Locked buffers..
		 */
		while(buffer_locked(bh))
		{
			/* Buffer is locked; skip it unless wait is
			   requested AND pass > 0. */
			if (!wait) 
				continue;
			wait_on_buffer (bh);
		}
		/*
		 *	Do the stuff
		 */
		bh->b_count++;
		ll_rw_block(WRITE, 1, &bh);
		bh->b_count--;
	}
	return;
}

void fsync_dev(dev)
kdev_t dev;
{
	sync_buffers(dev, 0);
	sync_supers(dev);
	sync_inodes(dev);
	sync_buffers(dev, 1);
}

void sync_dev(dev)
kdev_t dev;
{
	sync_buffers(dev, 0);
	sync_supers(dev);
	sync_inodes(dev);
	sync_buffers(dev, 0);
}

int sys_sync()
{
	fsync_dev(0);
	return 0;
}

void invalidate_buffers(dev)
kdev_t dev;
{
	struct buffer_head * bh;
	
	for(bh = bh_chain; bh!=NULL; bh=bh->b_next)
	{
		if (bh->b_dev != dev)
			continue;
		wait_on_buffer(bh);
		if (bh->b_dev != dev)
			continue;
		if (bh->b_count)
			continue;
		clear_bit(BH_Protected, &bh->b_state);
		clear_bit(BH_Uptodate, &bh->b_state);
		clear_bit(BH_Dirty, &bh->b_state);
		clear_bit(BH_Req, &bh->b_state);
	}
}

static struct buffer_head * find_buffer(dev,block,size)
kdev_t dev;
unsigned long block;
int size;
{		
	struct buffer_head * tmp;

	for (tmp = bh_chain ; tmp != NULL ; tmp = tmp->b_next)
	{
		if (tmp->b_blocknr == block && tmp->b_dev == dev)
				return tmp;
	}
	return NULL;
}

static struct buffer_head *get_free_buffer()
{
	do
	{
		struct buffer_head *bh=bh_lru;
		while(bh)
		{
			if(bh->b_count==0 && !buffer_dirty(bh) && !buffer_locked(bh))
			{
				put_last_lru(bh);
				return bh;
			}
			bh=bh->b_next_lru;
		}
		fsync_dev(0);
		sleep_on(&bufwait);
	}
	while(1);
}
	
struct buffer_head * get_hash_table(dev,block,size)
kdev_t dev;
unsigned long block;
int size;
{
	struct buffer_head * bh;

	for (;;) {
		if (!(bh=find_buffer(dev,block,size)))
			return NULL;
		bh->b_count++;
		wait_on_buffer(bh);
		if (bh->b_dev == dev && bh->b_blocknr == block
		                             && bh->b_size == size)
			return bh;
		bh->b_count--;
	}
}

/*
 * Ok, this is getblk, and it isn't very clear, again to hinder
 * race-conditions. Most of the code is seldom used, (ie repeating),
 * so it should be much more efficient than it looks.
 *
 * The algorithm is changed: hopefully better, and an elusive bug removed.
 *
 * 14.02.92: changed it to sync dirty buffers a bit: better performance
 * when the filesystem starts to get full of dirty blocks (I hope).
 */

struct buffer_head * getblk(dev,block,size)
kdev_t dev;
unsigned long block;
int size;
{
	struct buffer_head * bh;

	/* If there are too many dirty buffers, we wake up the update process
	   now so as to ensure that there are still clean buffers available
	   for user processes to use (and dirty) */
repeat:
	bh = get_hash_table(dev, block, size);
	if (bh) {
		if (!buffer_dirty(bh)) {
			if (buffer_uptodate(bh))
				 put_last_lru(bh);
		}
		set_bit(BH_Touched, &bh->b_state);
		return bh;
	}

	if (find_buffer(dev,block,size))
		 goto repeat;
	/*
	 *	Create a buffer for this job.
	 */
	bh = get_free_buffer();

/* OK, FINALLY we know that this buffer is the only one of its kind, */
/* and that it's unused (b_count=0), unlocked (buffer_locked=0), and clean */
	bh->b_count=1;
	bh->b_state=(1<<BH_Touched);
	bh->b_dev=dev;
	bh->b_blocknr=block;
	bh->b_next=bh_chain;
	bh->b_size=size;
	bh_chain=bh;
	return bh;
}

/*
 * Release a buffer head
 */

void __brelse(buf)
struct buffer_head * buf;
{
	wait_on_buffer(buf);

	if (buf->b_count) 
	{
		buf->b_count--;
		wake_up(&bufwait);
		return;
	}
	printk("VFS: brelse: Trying to free free buffer\n");
}

/*
 * bforget() is like brelse(), except it removes the buffer 
 * data validity.
 */
 
void __bforget(buf)
struct buffer_head * buf;
{
	wait_on_buffer(buf);
	mark_buffer_clean(buf);
	clear_bit(BH_Protected, &buf->b_state);
	buf->b_count--;
	buf->b_dev = NODEV;
	wake_up(&bufwait);
}

/*
 * bread() reads a specified block and returns the buffer that contains
 * it. It returns NULL if the block was unreadable.
 */

struct buffer_head * bread(dev,block,size)
kdev_t dev;
unsigned long block;
int size;
{
	struct buffer_head * bh;

	if (!(bh = getblk(dev, block, size))) {
		printk("VFS: bread: READ error on device %s\n",
			kdevname(dev));
		return NULL;
	}
	if (buffer_uptodate(bh))
		return bh;
	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);
	if (buffer_uptodate(bh))
		return bh;
	brelse(bh);
	return NULL;
}

/*
 * Ok, breada can be used as bread, but additionally mark the next block
 * for reading too.
 */

struct buffer_head * breada(dev,block,bufsize,
	pos, filesize)
kdev_t dev;
unsigned long block;
int bufsize;
unsigned int pos;
unsigned int filesize;
{
	struct buffer_head * bh, *bha;
	int i, j;

	if (pos >= filesize)
		return NULL;

	if (block < 0 || !(bh = getblk(dev,block,bufsize)))
		return NULL;

	if (buffer_uptodate(bh))
		return bh;

	bha = getblk(dev,block+1,bufsize);
	if (buffer_uptodate(bha)) 
	{
		brelse(bha);
		bha=NULL;
	}
	else
	{
		/* Request the read for these buffers, and then release them */
		ll_rw_block(READ, 1, &bha);
		brelse(bha);
	}
	/* Wait for this buffer, and then continue on */
	wait_on_buffer(bh);
	if (buffer_uptodate(bh))
		return bh;
	brelse(bh);
	return NULL;
}

void mark_buffer_uptodate(bh,on)
struct buffer_head * bh;
int on;
{
	int flags;
	save_flags(flags);
	cli();
	if (on) 
		bh->b_state&=~(1<<BH_Uptodate);
	else
		bh->b_state|=(1<<BH_Uptodate);
	restore_flags(flags);
}

void unlock_buffer(bh)
struct buffer_head * bh;
{
	struct buffer_head *tmp;
	clear_bit(BH_Lock, &bh->b_state);
	wake_up(&bh->b_wait);
}

void buffer_init()
{
	struct buffer_head *bh=&buffers[0];
	int i;
	for(i=0;i<NR_BUFFERS;i++)
	{
		if(i==0)
		{
			bh_chain=bh;
			bh_lru=bh;
			bh->b_prev=NULL;
			bh->b_prev_lru=NULL;
		}
		else
		{
			bh->b_prev_lru=bh-1;
			bh->b_prev=bh-1;
		}
		if(i==NR_BUFFERS-1)
		{
			bh->b_next_lru=NULL;
			bh->b_next=NULL;
			bh_llru=bh;
		}
		else
		{
			bh->b_next_lru=bh+1;
			bh->b_next=bh+1;
		}
		bh++;
	}
}
